@@ -39,6 +39,10 @@ typedef unsigned long uLong;
 #define M_PI 3.14159265358979323846
 #endif
 
+#ifdef HAVE_ZLIB
+#define COMPRESS_CHUNK_SIZE 10485760 /* 10 MiB */
+#endif
+
 #ifndef min
 #define min(a, b) (((a) < (b)) ? (a) : (b))
 #endif
@@ -120,7 +124,10 @@ typedef struct PDF_image_t
 {
   long object;
   int width, height;
-  int *data;
+  Byte *rgb;
+  uLong rgb_length;
+  Byte *alpha;
+  uLong alpha_length;
 } PDF_image;
 
 typedef struct PDF_page_t
@@ -248,11 +255,127 @@ static double xfac[4] = {0, 0, -0.5, -1};
 
 static double yfac[6] = {0, -1.2, -1, -0.5, 0, 0.2};
 
+static double cx[4][3] = {{0.5523, 1, 1}, {1, 0.5523, 0}, {-0.5523, -1, -1}, {-1, -0.5523, 0}};
+
+static double cy[4][3] = {{-1, -0.5523, 0}, {0.5523, 1, 1}, {1, 0.5523, 0}, {-0.5523, -1, -1}};
+
 static void fill_routine(int n, double *px, double *py, int tnr);
 
 static char buf_array[NO_OF_BUFS][20];
 static int current_buf = 0;
 
+#ifdef HAVE_ZLIB
+static int swap_index(uLong index, uLong width, uLong height, uLong dimx, int swapx, int swapy)
+{
+  int row = index / width;
+  int col = index % width;
+  if (swapx) col = width - col - 1;
+  if (swapy) row = height - row - 1;
+  return row * dimx + col;
+}
+
+int compress_chunkwise2d(Byte **out_buffer, uLong *out_length, const Byte *in_buffer, uLong width, uLong height,
+                         uLong dimx, int swapx, int swapy, const Byte *byte_drop_mask, uLong byte_drop_mask_length)
+{
+  z_stream strm;
+  Byte *compressed_buffer = NULL, *enlarged_buffer = NULL, *in_chunk = NULL;
+  uLong compressed_length = COMPRESS_CHUNK_SIZE, in_chunk_length;
+  uLong in_buffer_pos = 0;
+  const uLong in_length = width * height;
+  const int use_input_chunk = (byte_drop_mask != NULL && byte_drop_mask_length > 0) || swapx || swapy || dimx != width;
+  int initialized_stream = 0;
+  int was_successful = 1;
+
+  strm.zalloc = Z_NULL;
+  strm.zfree = Z_NULL;
+  strm.opaque = Z_NULL;
+  int ret = deflateInit(&strm, Z_DEFAULT_COMPRESSION);
+  if (ret != Z_OK)
+    {
+      was_successful = 0;
+      goto cleanup;
+    }
+  strm.avail_out = 0;
+  initialized_stream = 1;
+
+  compressed_buffer = malloc(compressed_length);
+  if (compressed_buffer == NULL)
+    {
+      was_successful = 0;
+      goto cleanup;
+    }
+  if (use_input_chunk)
+    {
+      in_chunk_length = COMPRESS_CHUNK_SIZE;
+      in_chunk = malloc(in_chunk_length);
+    }
+  else
+    {
+      strm.avail_in = in_length;
+      strm.next_in = (Byte *)in_buffer;
+    }
+
+  do
+    {
+      if (use_input_chunk)
+        {
+          uLong i;
+          for (i = 0; i < in_chunk_length; i++)
+            {
+              while (byte_drop_mask[swap_index(in_buffer_pos, width, height, dimx, swapx, swapy) %
+                                    byte_drop_mask_length] == 0)
+                {
+                  ++in_buffer_pos;
+                }
+              if (in_buffer_pos >= in_length) break;
+              in_chunk[i] = in_buffer[swap_index(in_buffer_pos, width, height, dimx, swapx, swapy)];
+              ++in_buffer_pos;
+            }
+          strm.avail_in = i;
+          strm.next_in = in_chunk;
+        }
+      while (1)
+        {
+          if (strm.avail_out == 0)
+            {
+              strm.avail_out = COMPRESS_CHUNK_SIZE;
+              strm.next_out = compressed_buffer + compressed_length - COMPRESS_CHUNK_SIZE;
+            }
+
+          ret = deflate(&strm, (use_input_chunk && in_buffer_pos < in_length) ? Z_NO_FLUSH : Z_FINISH);
+
+          if (ret == Z_STREAM_END || strm.avail_out > 0)
+            break;
+          else if (ret == Z_STREAM_ERROR)
+            {
+              was_successful = 0;
+              goto cleanup;
+            }
+
+          compressed_length += COMPRESS_CHUNK_SIZE;
+          if ((enlarged_buffer = realloc(compressed_buffer, compressed_length)) == NULL)
+            {
+              was_successful = 0;
+              goto cleanup;
+            }
+          compressed_buffer = enlarged_buffer;
+        };
+    }
+  while (use_input_chunk && in_buffer_pos < in_length);
+
+  *out_buffer = compressed_buffer;
+  *out_length = compressed_length - strm.avail_out;
+
+cleanup:
+  if (initialized_stream) deflateEnd(&strm);
+  if (!was_successful) free(compressed_buffer);
+  free(in_chunk);
+
+  return was_successful;
+}
+
+#endif
+
 static const char *pdf_double(double f)
 {
   char *buf = buf_array[(current_buf++) % NO_OF_BUFS];
@@ -365,9 +488,13 @@ static void pdf_open(int fd)
   p->preview_fix = (char *)gks_getenv("GKS_PDF_PREVIEW_FIX") != NULL ? 1 : 0;
 }
 
-static PDF_image *pdf_image(PDF *p, int width, int height)
+static PDF_image *pdf_image(PDF *p, int width, int height, int dimx, int swapx, int swapy, int *image_rgba)
 {
   PDF_image *image;
+#ifdef HAVE_ZLIB
+  const Byte alpha_mask[] = {0x00, 0x00, 0x00, 0xff};
+  const Byte rgb_mask[] = {0xff, 0xff, 0xff, 0x00};
+#endif
 
   if (p->images + 1 >= MAX_IMAGES)
     {
@@ -380,7 +507,48 @@ static PDF_image *pdf_image(PDF *p, int width, int height)
   image->object = pdf_alloc_id(p);
   image->width = width;
   image->height = height;
-  image->data = (int *)pdf_calloc(width * height, sizeof(int));
+
+#ifdef HAVE_ZLIB
+  if (p->compress)
+    {
+      if (gkss->debug)
+        {
+          fprintf(stdout, "[DEBUG:GKS] ZLIB version %s\n", zlibVersion());
+        }
+
+      compress_chunkwise2d(&image->rgb, &image->rgb_length, (Byte *)image_rgba, width * sizeof(int), height,
+                           dimx * sizeof(int), swapx, swapy, rgb_mask, sizeof(rgb_mask));
+      compress_chunkwise2d(&image->alpha, &image->alpha_length, (Byte *)image_rgba, width * sizeof(int), height,
+                           dimx * sizeof(int), swapx, swapy, alpha_mask, sizeof(alpha_mask));
+    }
+  else
+#endif
+    {
+      int i, j, iy, ix;
+
+      if (gkss->debug)
+        {
+          if (p->compress) fprintf(stdout, "[DEBUG:GKS] missing ZLIB compression library\n");
+        }
+
+      image->alpha = (Byte *)pdf_calloc(width * height, sizeof(Byte));
+      image->rgb = (Byte *)pdf_calloc(width * height, 3 * sizeof(Byte));
+      image->alpha_length = width * height;
+      image->rgb_length = width * height * 3 * sizeof(Byte);
+
+      for (j = 0; j < height; j++)
+        {
+          iy = swapy ? height - 1 - j : j;
+          for (i = 0; i < width; i++)
+            {
+              ix = swapx ? width - 1 - i : i;
+              image->rgb[3 * (j * width + i) + 0] = (image_rgba[iy * dimx + ix] & 0xff);
+              image->rgb[3 * (j * width + i) + 1] = (image_rgba[iy * dimx + ix] & 0xff00) >> 8;
+              image->rgb[3 * (j * width + i) + 2] = (image_rgba[iy * dimx + ix] & 0xff0000) >> 16;
+              image->alpha[j * width + i] = (image_rgba[iy * dimx + ix] & 0xff000000) >> 24;
+            }
+        }
+    }
 
   return image;
 }
@@ -418,7 +586,8 @@ static void pdf_close(PDF *p)
   struct tm ltime;
   long start_xref;
   int count, object, font, pattern;
-  int image, width, height, length, *rgba, alpha;
+  int image, width, height, length;
+  Byte *rgb, *alpha;
   Byte red, green, blue, data[3];
   int mask_id, filter_id, i;
   stroke_data_t s;
@@ -477,6 +646,7 @@ static void pdf_close(PDF *p)
   pdf_stream(p);
   pdf_printf(p->stream, "000000ffffff\n");
   pdf_endstream(p);
+  pdf_endobj(p);
   for (pattern = 0; pattern < PATTERNS; pattern++)
     {
       if (p->have_pattern[pattern])
@@ -521,6 +691,7 @@ static void pdf_close(PDF *p)
 
   for (count = 0; count < p->current_page; count++)
     {
+      int alpha;
       PDF_page *page = p->page[count];
 
       pdf_obj(p, page->object);
@@ -565,37 +736,7 @@ static void pdf_close(PDF *p)
       p->content = page->stream;
       pdf_obj(p, page->contents);
       pdf_dict(p);
-
-#ifdef HAVE_ZLIB
-      if (p->compress)
-        {
-          Byte *buffer;
-          uLong length;
-          int err;
-
-          length = p->content->length + 1024;
-          buffer = (Byte *)pdf_calloc((int)length, 1);
-          if ((err = compress(buffer, &length, p->content->buffer, p->content->length)) != Z_OK)
-            {
-              gks_perror("compression failed (err=%d)", err);
-              exit(-1);
-            }
-          free(p->content->buffer);
-
-          p->content->buffer = buffer;
-          p->content->size = p->content->length = length;
-          pdf_printf(p->stream, "/Length %ld\n", p->content->length);
-          pdf_printf(p->stream, "/Filter [/FlateDecode]\n");
-          buffer[p->content->length++] = '\n';
-        }
-      else
-        {
-          pdf_printf(p->stream, "/Length %ld\n", p->content->length);
-        }
-#else
       pdf_printf(p->stream, "/Length %ld\n", p->content->length);
-#endif
-
       pdf_enddict(p);
       pdf_stream(p);
       pdf_memcpy(p->stream, (char *)p->content->buffer, p->content->length);
@@ -648,8 +789,8 @@ static void pdf_close(PDF *p)
     {
       width = p->image[image]->width;
       height = p->image[image]->height;
-      length = width * height;
-      rgba = p->image[image]->data;
+      rgb = p->image[image]->rgb;
+      alpha = p->image[image]->alpha;
 
       mask_id = pdf_alloc_id(p);
       pdf_obj(p, mask_id);
@@ -660,22 +801,18 @@ static void pdf_close(PDF *p)
       pdf_printf(p->stream, "/ColorSpace /DeviceGray\n");
       pdf_printf(p->stream, "/Height %d\n", height);
       pdf_printf(p->stream, "/Width %d\n", width);
-      pdf_printf(p->stream, "/Length %d\n", length);
+      pdf_printf(p->stream, "/Length %d\n", p->image[image]->alpha_length);
+#ifdef HAVE_ZLIB
+      if (p->compress) pdf_printf(p->stream, "/Filter /FlateDecode\n");
+#endif
       pdf_enddict(p);
 
       pdf_stream(p);
-      for (i = 0; i < length; i++)
-        {
-          alpha = (*rgba & 0xff000000) >> 24;
-          rgba++;
-          pdf_memcpy(p->stream, (char *)&alpha, 1);
-        }
+      pdf_memcpy(p->stream, (char *)alpha, p->image[image]->alpha_length);
       pdf_printf(p->stream, "\n");
       pdf_endstream(p);
       pdf_endobj(p);
 
-      rgba = p->image[image]->data;
-
       pdf_obj(p, p->image[image]->object);
       pdf_dict(p);
       pdf_printf(p->stream, "/Type /XObject\n");
@@ -685,21 +822,14 @@ static void pdf_close(PDF *p)
       pdf_printf(p->stream, "/Height %d\n", height);
       pdf_printf(p->stream, "/Width %d\n", width);
       pdf_printf(p->stream, "/SMask %d 0 R\n", mask_id);
-      pdf_printf(p->stream, "/Length %d\n", length * 3);
+      pdf_printf(p->stream, "/Length %d\n", p->image[image]->rgb_length);
+#ifdef HAVE_ZLIB
+      if (p->compress) pdf_printf(p->stream, "/Filter /FlateDecode\n");
+#endif
       pdf_enddict(p);
 
       pdf_stream(p);
-      for (i = 0; i < length; i++)
-        {
-          red = (*rgba & 0xff);
-          green = (*rgba & 0xff00) >> 8;
-          blue = (*rgba & 0xff0000) >> 16;
-          rgba++;
-          data[0] = (Byte)red;
-          data[1] = (Byte)green;
-          data[2] = (Byte)blue;
-          pdf_memcpy(p->stream, (char *)data, 3);
-        }
+      pdf_memcpy(p->stream, (char *)rgb, p->image[image]->rgb_length);
       pdf_printf(p->stream, "\n");
       pdf_endstream(p);
       pdf_endobj(p);
@@ -774,10 +904,9 @@ static void set_xform(void)
 
   p->width = nint(p->a * (p->window[1] - p->window[0]));
   p->height = nint(p->c * (p->window[3] - p->window[2]));
-  if (gkss->resize_behaviour == GKS_K_RESIZE)
-    {
-      p->nominal_size = min(p->width, p->height) / 500.0;
-    }
+
+  p->nominal_size = min(p->width, p->height) / 500.0;
+  if (gkss->nominal_size > 0) p->nominal_size *= gkss->nominal_size;
 }
 
 static void seg_xform(double *x, double *y)
@@ -834,7 +963,7 @@ static void create_patterns(void)
       gks_inq_pattern_array(pattern, parray);
       for (j = 0, k = 1; j < 16; j += 2)
         {
-          snprintf(bitmap[i] + j, 17, "%02x", parray[k]);
+          snprintf(bitmap[i] + j, 17 - j, "%02x", parray[k]);
           if (++k > *parray) k = 1;
         }
       bitmap[i][16] = '\0';
@@ -865,7 +994,8 @@ static void open_ws(int fd, int wstype)
   p->viewport[0] = p->viewport[2] = 0;
   p->viewport[1] = p->viewport[3] = 0.1984;
   p->width = p->height = 558;
-  p->nominal_size = 558 / 500.0;
+  p->nominal_size = min(p->width, p->height) / 500.0;
+  if (gkss->nominal_size > 0) p->nominal_size *= gkss->nominal_size;
 
   p->empty = 1;
 
@@ -876,17 +1006,98 @@ static void open_ws(int fd, int wstype)
   pdf_open(fd);
 }
 
-static void set_clip(double *clrt)
+static void arc(double x, double y, double w, double h, double a1, double a2)
 {
-  double x0, x1, y0, y1;
+  double bcp, cos_a1, cos_a2, sin_a1, sin_a2;
+
+  a1 = a1 * M_PI / 180;
+  a2 = a2 * M_PI / 180;
+
+  bcp = (4.0 / 3 * (1 - cos(0.5 * (a2 - a1))) / sin(0.5 * (a2 - a1)));
+
+  sin_a1 = sin(a1);
+  sin_a2 = sin(a2);
+  cos_a1 = cos(a1);
+  cos_a2 = cos(a2);
+
+  pdf_printf(p->content, "%.2f %.2f %.2f %.2f %.2f %.2f c\n", x + w * (cos_a1 - bcp * sin_a1),
+             y + h * (sin_a1 + bcp * cos_a1), x + w * (cos_a2 + bcp * sin_a2), y + h * (sin_a2 - bcp * cos_a2),
+             x + w * cos_a2, y + h * sin_a2);
+}
+
+static void draw_arc(double x, double y, double w, double h, double a1, double a2)
+{
+  if (a1 == a2) return;
+
+  while (fabs(a2 - a1) > 90 + 0.1)
+    {
+      if (a2 > a1)
+        {
+          arc(x, y, w, h, a1, a1 + 90);
+          a1 += 90;
+        }
+      else
+        {
+          arc(x, y, w, h, a1, a1 - 90);
+          a1 -= 90;
+        }
+    }
+
+  if (a1 != a2) arc(x, y, w, h, a1, a2);
+}
+
+static void set_clip_rect(int tnr)
+{
+  double *clrt, x0, x1, y0, y1;
+  int curve, i;
+  double x, y, xr, yr;
+
+  if (gkss->clip_tnr != 0)
+    clrt = gkss->viewport[gkss->clip_tnr];
+  else if (gkss->clip == GKS_K_CLIP)
+    clrt = gkss->viewport[tnr];
+  else
+    clrt = gkss->viewport[0];
 
   NDC_to_DC(clrt[0], clrt[2], x0, y0);
   NDC_to_DC(clrt[1], clrt[3], x1, y1);
 
-  pdf_moveto(p, x0, y0);
-  pdf_lineto(p, x1, y0);
-  pdf_lineto(p, x1, y1);
-  pdf_lineto(p, x0, y1);
+  if (gkss->clip_region == GKS_K_REGION_ELLIPSE)
+    {
+      x = 0.5 * (x0 + x1);
+      y = 0.5 * (y0 + y1);
+      xr = 0.5 * (x1 - x0);
+      yr = 0.5 * (y1 - y0);
+      if (gkss->clip_start_angle > 0 || gkss->clip_end_angle < 360)
+        {
+          double w, h;
+          w = xr;
+          h = yr;
+          pdf_moveto(p, x + w * cos(gkss->clip_start_angle * M_PI / 180),
+                     y + h * sin(gkss->clip_start_angle * M_PI / 180));
+          draw_arc(x, y, w, h, gkss->clip_start_angle, gkss->clip_end_angle);
+          pdf_lineto(p, x, y);
+        }
+      else
+        {
+          pdf_moveto(p, x - xr * cx[3][2], y - yr * cy[3][2]);
+          for (curve = 0; curve < 4; curve++)
+            {
+              for (i = 0; i < 3; i++)
+                {
+                  pdf_point(p, x - xr * cx[curve][i], y - yr * cy[curve][i]);
+                }
+              pdf_curveto(p);
+            }
+        }
+    }
+  else
+    {
+      pdf_moveto(p, x0, y0);
+      pdf_lineto(p, x1, y0);
+      pdf_lineto(p, x1, y1);
+      pdf_lineto(p, x0, y1);
+    }
   pdf_closepath(p);
   pdf_clip(p);
 }
@@ -912,7 +1123,6 @@ static void begin_page(void)
 {
   init_context();
   pdf_page(p, p->width, p->height);
-  set_clip(p->window);
   p->empty = 0;
 }
 
@@ -1023,20 +1233,14 @@ static void polyline(int n, double *px, double *py)
   set_transparency(p->alpha);
   set_color(ln_color);
 
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_save(p);
-      set_clip(gkss->viewport[gkss->clip_tnr]);
-    }
+  pdf_save(p);
+  set_clip_rect(gkss->cntnr);
 
   gks_set_dev_xform(gkss, p->window, p->viewport);
   gks_emul_polyline(n, px, py, ln_type, gkss->cntnr, move, draw);
   stroke();
 
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_restore(p);
-    }
+  pdf_restore(p);
 }
 
 static void draw_marker(double xn, double yn, int mtype, double mscale, int mcolor)
@@ -1047,10 +1251,6 @@ static void draw_marker(double xn, double yn, int mtype, double mscale, int mcol
 
 #include "marker.h"
 
-  static double cx[4][3] = {{0.5523, 1, 1}, {1, 0.5523, 0}, {-0.5523, -1, -1}, {-1, -0.5523, 0}};
-
-  static double cy[4][3] = {{-1, -0.5523, 0}, {0.5523, 1, 1}, {1, 0.5523, 0}, {-0.5523, -1, -1}};
-
   mscale *= p->nominal_size;
   r = 3 * mscale;
   scale = 0.01 * mscale / 3.0;
@@ -1240,18 +1440,12 @@ static void polymarker(int n, double *px, double *py)
   set_linetype(GKS_K_LINETYPE_SOLID, 1.0);
   set_transparency(p->alpha);
 
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_save(p);
-      set_clip(gkss->viewport[gkss->clip_tnr]);
-    }
+  pdf_save(p);
+  set_clip_rect(gkss->cntnr);
 
   marker_routine(n, px, py, mk_type, mk_size, mk_color);
 
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_restore(p);
-    }
+  pdf_restore(p);
 }
 
 static void set_font(int font)
@@ -1371,11 +1565,8 @@ static void text(double px, double py, int nchars, char *chars)
 
   if (tx_prec != GKS_K_TEXT_PRECISION_STROKE) set_font(tx_font);
 
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_save(p);
-      set_clip(gkss->viewport[gkss->clip_tnr]);
-    }
+  pdf_save(p);
+  set_clip_rect(gkss->cntnr);
 
   if (tx_prec == GKS_K_TEXT_PRECISION_STRING)
     {
@@ -1389,10 +1580,7 @@ static void text(double px, double py, int nchars, char *chars)
       gks_emul_text(px, py, nchars, chars, line_routine, fill_routine);
     }
 
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_restore(p);
-    }
+  pdf_restore(p);
 }
 
 static void fill_routine(int n, double *px, double *py, int tnr)
@@ -1465,18 +1653,12 @@ static void fillarea(int n, double *px, double *py)
       set_transparency(p->alpha);
       set_color(fl_color);
 
-      if (gkss->clip_tnr != 0)
-        {
-          pdf_save(p);
-          set_clip(gkss->viewport[gkss->clip_tnr]);
-        }
+      pdf_save(p);
+      set_clip_rect(gkss->cntnr);
 
       line_routine(n, px, py, DrawBorder, gkss->cntnr);
 
-      if (gkss->clip_tnr != 0)
-        {
-          pdf_restore(p);
-        }
+      pdf_restore(p);
     }
   else if (fl_inter == GKS_K_INTSTYLE_SOLID)
     {
@@ -1484,14 +1666,7 @@ static void fillarea(int n, double *px, double *py)
       set_fillcolor(fl_color);
 
       pdf_save(p);
-      if (gkss->clip_tnr != 0)
-        {
-          set_clip(gkss->viewport[gkss->clip_tnr]);
-        }
-      else
-        {
-          set_clip(gkss->viewport[gkss->clip == GKS_K_CLIP ? gkss->cntnr : 0]);
-        }
+      set_clip_rect(gkss->cntnr);
 
       fill_routine(n, px, py, gkss->cntnr);
 
@@ -1507,14 +1682,7 @@ static void fillarea(int n, double *px, double *py)
       p->pattern = fl_style;
 
       pdf_save(p);
-      if (gkss->clip_tnr != 0)
-        {
-          set_clip(gkss->viewport[gkss->clip_tnr]);
-        }
-      else
-        {
-          set_clip(gkss->viewport[gkss->clip == GKS_K_CLIP ? gkss->cntnr : 0]);
-        }
+      set_clip_rect(gkss->cntnr);
 
       fill_routine(n, px, py, gkss->cntnr);
 
@@ -1544,7 +1712,6 @@ static void cellarray(double xmin, double xmax, double ymin, double ymax, int dx
   PDF_image *image;
   int swapx, swapy, count, chars_per_line;
   unsigned char data[3];
-  int have_alpha;
 
   WC_to_NDC(xmin, ymax, gkss->cntnr, x1, y1);
   seg_xform(&x1, &y1);
@@ -1566,26 +1733,7 @@ static void cellarray(double xmin, double xmax, double ymin, double ymax, int dx
   set_transparency(p->alpha);
 
   pdf_save(p);
-  if (gkss->clip_tnr != 0)
-    {
-      set_clip(gkss->viewport[gkss->clip_tnr]);
-    }
-  else
-    {
-      set_clip(gkss->viewport[gkss->clip == GKS_K_CLIP ? gkss->cntnr : 0]);
-    }
-
-  have_alpha = 0;
-  if (true_color)
-    {
-      for (j = 0; j < dy; j++)
-        for (i = 0; i < dx; i++)
-          if ((colia[j * dimx + i] & 0xff000000) != 0xff000000)
-            {
-              have_alpha = 1;
-              break;
-            }
-    }
+  set_clip_rect(gkss->cntnr);
 
   if (p->preview_fix)
     {
@@ -1623,23 +1771,13 @@ static void cellarray(double xmin, double xmax, double ymin, double ymax, int dx
             }
         }
     }
-  else if (true_color && have_alpha)
+  else if (true_color)
     {
       pdf_printf(p->content, "%d 0 0 %d %d %d cm\n", width, height, x, y);
 
-      image = pdf_image(p, dx, dy);
+      image = pdf_image(p, dx, dy, dimx, swapx, swapy, colia);
       p->image[p->images++] = image;
 
-      for (j = 0; j < dy; j++)
-        {
-          iy = swapy ? dy - 1 - j : j;
-          for (i = 0; i < dx; i++)
-            {
-              ix = swapx ? dx - 1 - i : i;
-              image->data[j * dx + i] = colia[iy * dimx + ix];
-            }
-        }
-
       pdf_printf(p->content, "/Im%d Do\n", p->images);
       p->page[p->current_page - 1]->last_image = p->images;
     }
@@ -1707,46 +1845,6 @@ static void to_DC(int n, double *x, double *y)
     }
 }
 
-static void arc(double x, double y, double w, double h, double a1, double a2)
-{
-  double bcp, cos_a1, cos_a2, sin_a1, sin_a2;
-
-  a1 = a1 * M_PI / 180;
-  a2 = a2 * M_PI / 180;
-
-  bcp = (4.0 / 3 * (1 - cos(0.5 * (a2 - a1))) / sin(0.5 * (a2 - a1)));
-
-  sin_a1 = sin(a1);
-  sin_a2 = sin(a2);
-  cos_a1 = cos(a1);
-  cos_a2 = cos(a2);
-
-  pdf_printf(p->content, "%.2f %.2f %.2f %.2f %.2f %.2f c\n", x + w * (cos_a1 - bcp * sin_a1),
-             y + h * (sin_a1 + bcp * cos_a1), x + w * (cos_a2 + bcp * sin_a2), y + h * (sin_a2 - bcp * cos_a2),
-             x + w * cos_a2, y + h * sin_a2);
-}
-
-static void draw_arc(double x, double y, double w, double h, double a1, double a2)
-{
-  if (a1 == a2) return;
-
-  while (fabs(a2 - a1) > 90)
-    {
-      if (a2 > a1)
-        {
-          arc(x, y, w, h, a1, a1 + 90);
-          a1 += 90;
-        }
-      else
-        {
-          arc(x, y, w, h, a1, a1 - 90);
-          a1 -= 90;
-        }
-    }
-
-  if (a1 != a2) arc(x, y, w, h, a1, a2);
-}
-
 static void draw_path(int n, double *px, double *py, int nc, int *codes)
 {
   int i, j;
@@ -2049,11 +2147,9 @@ static void fill_polygons(int n, double *px, double *py, int nply, int *ply)
 
 static void gdp(int n, double *px, double *py, int primid, int nc, int *codes)
 {
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_save(p);
-      set_clip(gkss->viewport[gkss->clip_tnr]);
-    }
+  pdf_save(p);
+  set_clip_rect(gkss->cntnr);
+
   set_linetype(GKS_K_LINETYPE_SOLID, 1.0);
   switch (primid)
     {
@@ -2076,19 +2172,12 @@ static void gdp(int n, double *px, double *py, int primid, int nc, int *codes)
       gks_perror("invalid drawing primitive ('%d')", primid);
       exit(1);
     }
-  if (gkss->clip_tnr != 0)
-    {
-      pdf_restore(p);
-    }
+
+  pdf_restore(p);
 }
 
-#ifndef EMSCRIPTEN
-void gks_drv_pdf(
-#else
-void gks_drv_js(
-#endif
-    int fctid, int dx, int dy, int dimx, int *ia, int lr1, double *r1, int lr2, double *r2, int lc, char *chars,
-    void **ptr)
+void gks_drv_pdf(int fctid, int dx, int dy, int dimx, int *ia, int lr1, double *r1, int lr2, double *r2, int lc,
+                 char *chars, void **ptr)
 {
   GKS_UNUSED(lr1);
   GKS_UNUSED(lr2);
@@ -2219,7 +2308,6 @@ void gks_drv_js(
 
       set_xform();
       init_norm_xform();
-      if (!p->empty) set_clip(p->window);
       break;
 
     case 55:
@@ -2233,6 +2321,12 @@ void gks_drv_js(
       init_norm_xform();
       break;
 
+    case 109:
+      /* set nominal size */
+      p->nominal_size = min(p->width, p->height) / 500.0;
+      if (gkss->nominal_size > 0) p->nominal_size *= gkss->nominal_size;
+      break;
+
     case 203:
       /* set transparency */
       p->alpha = (int)(r1[0] * 255.0);
